# Ruby library for parsing Nmap's XML output
#
# http://rubynmap.sourceforge.net
#
# Author: Kris Katterjohn <katterjohn@gmail.com>
#
# Copyright (c) 2007, 2008 Kris Katterjohn
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# $Id: nmapparser.rb 34 2008-01-04 21:56:08Z kjak $
# https://rubynmap.svn.sourceforge.net/svnroot/rubynmap

require 'rexml/document'

# Pretty much just holds the big Parser class.
module Nmap
end

# == What is this library for?
#
# This library provides a Ruby interface to Nmap's scan data.  It can run Nmap
# and parse it's XML output directly from the scan, parse a file containing the
# XML data from a separate scan, or parse a String of XML data from a scan.
# This information is presented in a (hopefully) easy to use fashion for storing
# and manipulating.
#
# The Nmap Security Scanner is a great program written and maintained by
# Fyodor <fyodor@insecure.org>.  It port scans, has service and operating
# system detection, it's own scripting engine and so much more.  One of it's
# many available output formats is XML, which allows machines to handle all
# of the information instead of us slowly sifting through tons of output!
#
# If you've ever used Anthony Persaud's Perl Nmap::Parser, you'll notice
# a lot of similarities and should pick up on this API quickly.  Just
# remember, it's not exactly the same!  There are more classes, many
# different methods, and blocks are used extensively (or available for use).
#
# == Module Hierarchy
#
#   Nmap::Parser
#   |
#   + ::Session             <- Scan session information
#   |
#   + ::Host                <- General host information
#      |
#      + ::Port             <- General port information
#      |  |
#      |  + ::Service       <- Port Service information
#      |
#      + ::Script           <- NSE Script information
#      |
#      + ::Times            <- Timimg information (rtt, etc)
#      |
#      + ::Traceroute       <- General Traceroute information
#      |  |
#      |  + ::Hop           <- Individual Hop information
#      |
#      + ::OS               <- OS Detection information
#         |
#         + ::OSClass       <- OS Class information
#         |
#         + ::OSMatch       <- OS Match information
#
# == Parsing XML data all ready available
#
# 	require 'nmapparser'
#
# 	parser = Nmap::Parser.new(xml) # String of XML data
#
# == Parsing from a file
#
#	require 'nmapparser'
#
#	parser = Nmap::Parser.parsefile("log.xml")
#
# == Scanning and Parsing
#
# This is the only parser option that requires Nmap to be available
#
#	require 'nmapparser'
#
#	parser = Nmap::Parser.parsescan("sudo nmap", "--script all 192.168.1.0/24")
#
# == Actually doing something
#
# After printing a little session information, this example will cycle
# through all of the up hosts, printing service information on the
# open TCP ports, along with the name and output of any scripts that
# ran (on the host and individual ports).
#
#	puts "Nmap args: #{parser.session.scan_args}"
#
#	puts "Runtime: #{parser.session.scan_time} seconds"
#
#	puts
#
#	parser.hosts("up") do |host|
# 		puts "** Host #{host.addr} is up **"
# 		puts
#
# 		host.tcp_ports("open") do |port|
# 			srv = port.service
#
# 			puts "Port ##{port.num} is open (#{port.reason})"
# 			puts "\tService: #{srv.name}" if srv.name
# 			puts "\tProduct: #{srv.product}" if srv.product
# 			puts "\tVersion: #{srv.product}" if srv.version
# 			puts
#
#			port.scripts.each do |script|
# 				puts "\tScript Name: #{script.id}"
# 				puts "\tScript Output: #{script.output}"
#				puts
# 			end
# 		end
#
#		if host.scripts.any?
#			puts "Host Scripts Run:"
#			puts
#
#			host.scripts.each do |script|
# 				puts "\tScript Name: #{script.id}"
# 				puts "\tScript Output: #{script.output}"
# 				puts
# 			end
# 		end
#
# 		puts
#	end
class Nmap::Parser
	attr_reader :rawxml
	attr_reader :session

	# Read and parse contents of the Nmap XML file +filename+
	#
	# Returns a new Nmap::Parser object, and passes it to a block if
	# one is given
	def self.parsefile(filename)
		begin
			p = new(File.read(filename))
		rescue
			raise "Can't read file \"#{filename}\": #{$!}"
		end

		yield p if block_given?

		p
	end

	# Runs "+nmap+ -d +args+ +targets+"; returns a new Nmap::Parser object,
	# and passes it to a block if one is given.
	#
	# +nmap+ is here to allow you to do things like:
	#
	# parser = Nmap::Parser.parsescan("sudo ./nmap", ....)
	#
	# and still make it easy for me to inject the options for XML
	# output and debugging.
	#
	# +args+ can't contain arguments like -oA, -oX, etc. as these can
	# interfere with Parser's processing.  If you need that other output,
	# just run Nmap yourself and pass -oX output to Parser via new.  Or,
	# you can use rawxml to grab the whole XML (as a String) and save it
	# to a different file.
	#
	# +targets+ is an array of targets which will be split and appended to
	# the command.  It's optional and only for convenience because you can
	# put any targets you want scanned in +args+.
	def self.parsescan(nmap, args, targets = [])
		raise "Output option (-o*) passed" if args =~ /[^-]-o/

		# Enable debugging, give us our XML output, pass args
		command = nmap + " -d -oX - " + args + " "

		command += targets.join(" ") if targets.any?

		begin
			output = IO.popen(command).read
		rescue Exception
			raise "Cannot execute Nmap: #{$!}"
		end

		if output.to_a[0] !~ /^<.*xml version.*>$/ or
		   output.to_a[-1] !~ /<\/nmaprun>$/
			raise "Nmap seemed to fail: Check arguments"
		end

		p = new(output)

		yield p if block_given?

		p
	end

	# Returns an array of Host objects, and passes them each to a block if
	# one is given
	#
	# If an argument is given, only hosts matching +status+ are given
	def hosts(status = "")
		shosts = []

		@hosts.each do |host|
			if status.empty? or host.status == status
				shosts << host
				yield host if block_given?
			end
		end

		shosts
	end

	# Returns a Host object for host the specified IP address
	def get_host(ip)
		@hosts.find do |host|
			host.addr == ip
		end
	end

	# Deletes host with the specified IP address
	def del_host(ip)
		@hosts.delete_if do |host|
			host.addr == ip
		end
	end

	alias delete_host del_host

	# Returns an array of IPs scanned, and passes them each to a
	# block if one is given
	#
	# If an argument is given, only hosts matching +status+ are given
	def get_ips(status = "")
		ips = []

		@hosts.each do |host|
			if status.empty? or host.status == status
				ips << host.addr
				yield host.addr if block_given?
			end
		end

		ips
	end

	private

	def initialize(xml)
		if not xml.is_a?String
			raise "Invalid input: you should be passing a String"
		end

		parse(xml)
	end

	def parse(xml)
		@rawxml = xml

		begin
			root = REXML::Document.new(xml).root
		rescue
			raise "Error parsing XML: #{$!}"
		end

		raise "Error in XML data" if root.nil?

		@session = Session.new(root)

		@hosts = []

		root.each_element('host') do |host|
			@hosts << Host.new(host)
		end
	end
end

# This holds session information, such as runtime, Nmap's arguments,
# and verbosity/debugging
class Nmap::Parser::Session
	attr_reader :scan_args, :nmap_version, :xml_version
	attr_reader :start_str, :start_time, :stop_str, :stop_time, :scan_time
	attr_reader :verbose, :debug
	attr_reader :min_parallelism, :max_parallelism

	alias verbosity verbose
	alias debugging debug

	# If no argument is given, returns the total number of services
	# that were scanned.
	#
	# It an argument is given, only returns the number of services
	# scanned for +type+
	def numservices(type = "")
		total = 0

		@scaninfo.each do |info|
			if type.empty?
				total += info.numservices.to_i
			elsif info.type == type
				return info.numservices
			end
		end

		total.to_s
	end

	# Returns the protocol associated with the specified scan +type+
	def scan_type_proto(type)
		@scaninfo.each do |info|
			return info.proto if info.type == type
		end

		nil
	end

	# Returns an array of all the scan types performed
	def scan_types
		types = []

		@scaninfo.each do |info|
			types << info.type
		end

		types
	end

	private

	def initialize(root)
		parse(root)
	end

	def parse(root)
		@scan_args = root.attributes['args']

		@nmap_version = root.attributes['version']

		@xml_version = root.attributes['xmloutputversion']

		@start_str = root.attributes['startstr']
		@start_time = root.attributes['start']

		@finish_str = root.elements['runstats/finished'].attributes['timestr']
		@finish_time = root.elements['runstats/finished'].attributes['time']

		@scan_time = (@finish_time.to_i - @start_time.to_i).to_s

		@verbose = root.elements['verbose'].attributes['level']
		@debug = root.elements['debugging'].attributes['level']

		if parallelism = root.elements['parallelism']
			@min_parallelism = parallelism.attributes['min']
			@max_parallelism = parallelism.attributes['max']
		end

		@scaninfo = []

		root.each_element('scaninfo') do |info|
			@scaninfo << ScanInfo.new(info)
		end
	end
end

class Nmap::Parser::Session::ScanInfo # :nodoc: all
	attr_reader :type, :proto, :numservices

	private

	def initialize(info)
		parse(info)
	end

	def parse(info)
		@type = info.attributes['type']
		@proto = info.attributes['protocol']
		@numservices = info.attributes['numservices']
	end
end

# This holds all of the information about a target host.
#
# Status, IP/MAC addresses, hostnames, all that.  Port information is
# available in this class; either accessed through here or directly
# from a Port object.
class Nmap::Parser::Host
	attr_reader :status, :reason
	attr_reader :addrtype, :ip4_addr, :ip6_addr
	attr_reader :mac_addr, :mac_vendor
	attr_reader :all_hostnames
	attr_reader :extraports
	attr_reader :os
	attr_reader :smurf
	attr_reader :tcpsequence_index, :tcpsequence_class
	attr_reader :tcpsequence_values, :tcpsequence_difficulty
	attr_reader :ipidsequence_class, :ipidsequence_values
	attr_reader :tcptssequence_class, :tcptssequence_values
	attr_reader :uptime_seconds, :uptime_lastboot
	attr_reader :traceroute
	attr_reader :distance
	attr_reader :times

	alias ipv4_addr ip4_addr
	alias ipv6_addr ip6_addr

	# Returns the IPv4 or IPv6 address of host
	def addr
		@ip4_addr or @ip6_addr
	end

	# Returns the first hostname, or nil if unavailable
	def hostname
		@all_hostnames[0]
	end

	# Returns an array of Port objects for each TCP port, and passes them
	# each to a block if one is given
	#
	# If an argument is given, only ports matching +state+ are given.
	# Note:  "open" will also match "open|filtered", but not vice versa.
	def tcp_ports(state = "")
		list = []

		@tcpPorts.invert.each_key do |port|
			list << port if state.empty? or port.state >= state
		end

		list.sort! do |x, y|
			x.num.to_i <=> y.num.to_i
		end


		if block_given?
			list.each do |port|
				yield port
			end
		end

		list
	end

	# Returns an array of TCP port numbers, and passes them each to a block
	# if one given
	#
	# If an argument is given, only ports matching +state+ are given.
	# Note:  "open" will also match "open|filtered", but not vice versa.
	def tcp_port_list(state = "")
		list = []

		@tcpPorts.invert.each_key do |port|
			list << port.num if state.empty? or port.state >= state
		end

		list.sort! do |x, y|
			x.to_i <=> y.to_i
		end

		if block_given?
			list.each do |port|
				yield port
			end
		end

		list
	end

	# Returns state reason of TCP port +port+
	def tcp_reason(port)
		return nil if @tcpPorts[port].nil?

		@tcpPorts[port].reason
	end

	# Returns list of Script objects for each script run on the specified
	# TCP port +port+, and passes them each to a block if given
	def tcp_scripts(port)
		return nil if @tcpPorts[port].nil?

		if block_given?
			@tcpPorts[port].scripts.each do |script|
				yield script
			end
		end

		@tcpPorts[port].scripts
	end

	# Returns the output of the specified script on the specified TCP port
	# +port+
	def tcp_script_output(port, name)
		return nil if @tcpPorts[port].nil?

		@tcpPorts[port].scripts.each do |script|
			return script.output if script.id == name
		end

		nil
	end

	# Returns a Port::Service object for TCP port +port+
	def tcp_service(port)
		return nil if @tcpPorts[port].nil?

		@tcpPorts[port].service
	end

	# Returns state of TCP port +port+
	def tcp_state(port)
		return nil if @tcpPorts[port].nil?

		@tcpPorts[port].state
	end

	# Returns an array of Port objects for each UDP port, and passes them
	# each to a block if one is given
	#
	# If an argument is given, only ports matching +state+ are given.
	# Note:  "open" will also match "open|filtered", but not vice versa.
	def udp_ports(state = "")
		list = []

		@udpPorts.invert.each_key do |port|
			list << port if state.empty? or port.state >= state
		end

		list.sort! do |x, y|
			x.num.to_i <=> y.num.to_i
		end

		if block_given?
			list.each do |port|
				yield port
			end
		end

		list
	end

	# Returns an array of UDP port numbers, and passes them each to a block
	# if one is given
	#
	# If an argument is given, only ports matching +state+ are given.
	# Note:  "open" will also match "open|filtered", but not vice versa.
	def udp_port_list(state = "")
		list = []

		@udpPorts.invert.each_key do |port|
			list << port.num if state.empty? or port.state >= state
		end

		list.sort! do |x, y|
			x.to_i <=> y.to_i
		end

		if block_given?
			list.each do |port|
				yield port
			end
		end

		list
	end

	# Returns state reason of UDP port +port+
	def udp_reason(port)
		return nil if @udpPorts[port].nil?

		@udpPorts[port].reason
	end

	# Returns list of Script objects for each script run on the specified
	# UDP port +port+, and passes them each to a block if given
	def udp_scripts(port)
		return nil if @udpPorts[port].nil?

		if block_given?
			@udpPorts[port].scripts.each do |script|
				yield script
			end
		end

		@udpPorts[port].scripts
	end

	# Returns the output of the specified script on the specified UDP port
	# +port+
	def udp_script_output(port, name)
		return nil if @udpPorts[port].nil?

		@udpPorts[port].scripts.each do |script|
			return script.output if script.id == name
		end

		nil
	end

	# Returns a Port::Service object for UDP port +port+
	def udp_service(port)
		return nil if @udpPorts[port].nil?

		@udpPorts[port].service
	end

	# Returns state of UDP port +port+
	def udp_state(port)
		return nil if @udpPorts[port].nil?

		@udpPorts[port].state
	end

	# Returns an array of Port objects for each IP protocol, and passes
	# them each to a block if one is given
	#
	# If an argument is given, only protocols matching +state+ are given.
	# Note:  "open" will also match "open|filtered", but not vice versa.
	def ip_protos(state = "")
		list = []

		@ipProtos.invert.each_key do |proto|
			list << proto if state.empty? or proto.state >= state
		end

		list.sort! do |x, y|
			x.num.to_i <=> y.num.to_i
		end


		if block_given?
			list.each do |proto|
				yield proto
			end
		end

		list
	end

	# Returns an array of IP protocol numbers, and passes them each to a
	# block if one given
	#
	# If an argument is given, only protocols matching +state+ are given.
	# Note:  "open" will also match "open|filtered", but not vice versa.
	def ip_proto_list(state = "")
		list = []

		@ipProtos.invert.each_key do |proto|
			list << proto.num if state.empty? or proto.state >= state
		end

		list.sort! do |x, y|
			x.to_i <=> y.to_i
		end

		if block_given?
			list.each do |proto|
				yield proto
			end
		end

		list
	end

	# Returns state reason of IP protocol +proto+
	def ip_reason(proto)
		return nil if @ipProtos[proto].nil?

		@ipProtos[proto].reason
	end

	# Returns a Port::Service object for IP protocol +proto+
	def ip_service(proto)
		return nil if @ipProtos[proto].nil?

		@ipProtos[proto].service
	end

	# Returns state of IP protocol +proto+
	def ip_state(proto)
		return nil if @ipProtos[proto].nil?

		@ipProtos[proto].state
	end

	# Returns list of Script objects for each host script run, and passes
	# them each to a block if given
	def scripts
		if block_given?
			@scripts.each do |script|
				yield script
			end
		end

		@scripts
	end

	# Returns the output of the specified script +name+, or nil if an
	# invalid name is specified
	def script_output(name)
		@scripts.each do |script|
			return script.output if script.id == name
		end

		nil
	end

	private

	def initialize(hostinfo)
		parse(hostinfo)
	end

	# Convert a string like '22-25,80,110-900' into
	# an array with all the uncondensed elements, e.g.
	# [ '22', '23', '24', '25, '80', '110', '111' ... ]
	def parsePortlist(ports)
		list = []

		ports.split(",").each do |set|
			pa = set.split("-")

			if pa.length > 1
				pa[0].upto(pa[1]) do |p|
					list << p
				end
			else
				list << portc
			end
		end

		list
	end

	def parseAddr(elem)
		case @addrtype = elem.attributes['addrtype']
		when "mac"
			@mac_addr = elem.attributes['addr']
			@mac_vendor = elem.attributes['vendor']
		when "ipv4"
			@ip4_addr = elem.attributes['addr']
		when "ipv6"
			@ip6_addr = elem.attributes['addr']
		end
	end

	def parseHostnames(elem)
		return nil if elem.nil?

		elem.each_element('hostname') do |name|
			@all_hostnames << name.attributes['name']
		end
	end

	def parsePorts(ports)
		ports.each_element('port') do |port|
			num = port.attributes['portid']
			proto = port.attributes['protocol']

			if proto == "tcp"
				@tcpPorts[num] = Port.new(port)
			elsif proto == "udp"
				@udpPorts[num] = Port.new(port)
			elsif proto == "ip"
				@ipProtos[num] = Port.new(port)
			end
		end
	end

	def parseScripts(scriptlist)
		return nil if scriptlist.nil?

		scriptlist.each_element('script') do |script|
			@scripts << Script.new(script)
		end
	end

	def tcpseq(seq)
		return nil if seq.nil?

		@tcpsequence_index = seq.attributes['index']
		@tcpsequence_class = seq.attributes['class']
		@tcpsequence_values = seq.attributes['values']
		@tcpsequence_difficulty = seq.attributes['difficulty']
	end

	def ipidseq(seq)
		return nil if seq.nil?

		@ipidsequence_class = seq.attributes['class']
		@ipidsequence_values = seq.attributes['values']
	end

	def tcptsseq(seq)
		return nil if seq.nil?

		@tcptssequence_class = seq.attributes['class']
		@tcptssequence_values = seq.attributes['values']
	end

	def uptime(time)
		return nil if time.nil?

		@uptime_seconds = time.attributes['seconds']
		@uptime_lastboot = time.attributes['lastboot']
	end

	def parse(host)
		status = host.elements['status']

		@status = status.attributes['state']

		@reason = status.attributes['reason']

		@os = OS.new(host.elements['os'])

		host.each_element('address') do |elem|
			parseAddr(elem)
		end

		@all_hostnames = []

		parseHostnames(host.elements['hostnames'])

		smurf = host.elements['smurf']
		@smurf = smurf.attributes['responses'] if smurf

		@tcpPorts = {}
		@udpPorts = {}
		@ipProtos = {}

		ports = host.elements['ports']

		parsePorts(ports) if ports

		@extraports = []

		if ports
			ports.each_element('extraports') do |e|
				@extraports << ExtraPorts.new(e)
			end

			@extraports.sort! do |x, y|
				x.count.to_i <=> y.count.to_i
			end
		end

		@scripts = []

		parseScripts(host.elements['hostscript'])

		if trace = host.elements['trace']
			@traceroute = Traceroute.new(trace)
		end

		tcpseq(host.elements['tcpsequence'])

		ipidseq(host.elements['ipidsequence'])

		tcptsseq(host.elements['tcptssequence'])

		uptime(host.elements['uptime'])

		distance = host.elements['distance']

		@distance = distance.attributes['value'] if distance

		@times = Times.new(host.elements['times'])
	end
end

# This holds information on the time statistics for this host
class Nmap::Parser::Host::Times
	attr_reader :srtt, :rttvar, :to

	private

	def initialize(times)
		parse(times)
	end

	def parse(times)
		return nil if times.nil?

		@srtt = times.attributes['srtt']
		@rttvar = times.attributes['rttvar']
		@to = times.attributes['to']
	end
end

# This holds the information about NSE scripts run against the target host and
# ports
class Nmap::Parser::Host::Script
	attr_reader :id, :output

	alias name id

	private

	def initialize(script)
		parse(script)
	end

	def parse(script)
		return nil if script.nil?

		@id = script.attributes['id']
		@output = script.attributes['output']
	end
end

# This holds the information about individual ports/protocols
class Nmap::Parser::Host::Port
	attr_reader :num, :state, :service, :scripts
	attr_reader :reason, :reason_ip, :reason_ttl

	private

	def initialize(portinfo)
		parse(portinfo)
	end

	def parse(port)
		@num = port.attributes['portid']

		state = port.elements['state']
		@state = state.attributes['state']
		@reason = state.attributes['reason']
		@reason_ttl = state.attributes['reason_ttl']
		@reason_ip = state.attributes['reason_ip']

		@service = Service.new(port)

		@scripts = []

		port.each_element('script') do |script|
			@scripts << Nmap::Parser::Host::Script.new(script)
		end
	end
end

# This holds the information about "extra ports", groups of ports which have
# the same state.  +count+ is the total amount of ports in this group, which
# are all in state +state+
#
# +reasons+ is an array of arrays, each of which are in the form of:
#
# [ <port count>, reason ]
#
# for each set of reasons.
class Nmap::Parser::Host::ExtraPorts
	attr_reader :count, :state, :reasons

	private

	def initialize(extraports)
		parse(extraports)
	end

	def parse(extraports)
		@count = extraports.attributes['count']
		@state = extraports.attributes['state']

		@reasons = []

		extraports.each_element('extrareasons') do |extra|
			ecount = extra.attributes['count']
			ereason = extra.attributes['reason']
			@reasons << [ ecount, ereason ]
		end
	end
end

# This holds information on a traceroute, with +hops+ being an array of Hop
# objects (responsive hops)
class Nmap::Parser::Host::Traceroute
	attr_reader :port, :proto, :hops

	private

	def initialize(trace)
		parse(trace)
	end

	def parse(trace)
		@port = trace.attributes['port']
		@proto = trace.attributes['proto']

		@hops = []

		trace.each_element('hop') do |hop|
			@hops << Hop.new(hop)
		end
	end
end

# This holds information on the individual hops of a traceroute
class Nmap::Parser::Host::Traceroute::Hop
	attr_reader :ttl, :rtt, :ipaddr, :host

	private

	def initialize(hop)
		parse(hop)
	end

	def parse(hop)
		@ttl = hop.attributes['ttl']
		@rtt = hop.attributes['rtt']
		@ipaddr = hop.attributes['ipaddr']
		@host = hop.attributes['host']
	end
end

# This holds the service information for ports
class Nmap::Parser::Host::Port::Service
	attr_reader :name, :product, :version, :method, :owner, :tunnel
	attr_reader :rpcnum, :lowver, :highver, :confidence, :proto
	attr_reader :extra, :devicetype, :ostype, :fingerprint

	alias extrainfo extra

	private

	def initialize(port)
		parse(port)
	end

	def parse(port)
		return nil if port.nil?

		service = port.elements['service']

		return nil if service.nil?

		@name = service.attributes['name']
		@product = service.attributes['product']
		@version = service.attributes['version']
		@method = service.attributes['method']
		owner = port.elements['owner']
		@owner = owner.attributes['name'] if owner
		@tunnel = service.attributes['tunnel']
		@rpcnum = service.attributes['rpcnum']
		@lowver = service.attributes['lowver']
		@highver = service.attributes['highver']
		@confidence = service.attributes['conf']
		@proto = service.attributes['proto']
		@extra = service.attributes['extrainfo']
		@devicetype = service.attributes['devicetype']
		@ostype = service.attributes['ostype']
		@fingerprint = service.attributes['servicefp']
	end
end

# This holds the OS information from OS Detection
class Nmap::Parser::Host::OS
	attr_reader :osclasses, :osmatches, :fingerprint

	# Returns the number of OS class records
	def class_count
		@osclasses.size
	end

	# Returns OS class accuracy of the first OS class record, or Nth record
	# as specified by +index+
	def class_accuracy(index = 0)
		return nil if @osclasses.empty?

		@osclasses[index].accuracy
	end

	# Returns OS family information of first OS class record, or Nth
	# record as specified by +index+
	def osfamily(index = 0)
		return nil if @osclasses.empty?

		@osclasses[index].osfamily
	end

	# Returns OS generation information of first OS class record, or Nth
	# record as specified by +index+
	def osgen(index = 0)
		return nil if @osclasses.empty?

		@osclasses[index].osgen
	end

	# Returns OS type information of the first OS class record, or Nth
	# record as specified by +index+
	def ostype(index = 0)
		return nil if @osclasses.empty?

		@osclasses[index].ostype
	end

	# Returns OS vendor information of the first OS class record, or Nth
	# record as specified by +index+
	def osvendor(index = 0)
		return nil if @osclasses.empty?

		@osclasses[index].osvendor
	end

	# Returns the number of OS match records
	def name_count
		@osmatches.size
	end

	# Returns name of first OS match record, or Nth record as specified by
	# +index+
	def name(index = 0)
		return nil if @osmatches.empty?

		@osmatches[index].name
	end

	# Returns OS name accuracy of the first OS match record, or Nth record
	# as specified by +index+
	def name_accuracy(index = 0)
		return nil if @osmatches.empty?

		@osmatches[index].accuracy
	end

	# Returns an array of names from all OS records
	def all_names
		names = []

		@osmatches.each do |match|
			names << match.name
			yield match.name if block_given?
		end

		names
	end

	# Returns the closed TCP port used for this OS Detection run
	def tcpport_closed
		return nil if @portused.nil?

		@portsused.each do |port|
			if port.proto == "tcp" and port.state == "closed"
				return port.num
			end
		end

		nil
	end

	# Returns the open TCP port used for this OS Detection run
	def tcpport_open
		return nil if @portused.nil?

		@portsused.each do |port|
			if port.proto == "tcp" and port.state == "open"
				return port.num
			end
		end

		nil
	end

	# Returns the closed UDP port used for this OS Detection run
	def udpport_closed
		return nil if @portused.nil?

		@portsused.each do |port|
			if port.proto == "udp" and port.state == "closed"
				return port.num
			end
		end

		nil
	end

	private

	def initialize(os)
		parse(os)
	end

	def parse(os)
		@portsused = []
		@osclasses = []
		@osmatches = []

		return nil if os.nil?

		os.each_element('portused') do |port|
			@portsused << Port.new(port)
		end

		os.each_element('osclass') do |osclass|
			@osclasses << OSClass.new(osclass)
		end

		os.each_element('osmatch') do |match|
			@osmatches << OSMatch.new(match)
		end

		fp = os.elements['osfingerprint']
		@fingerprint = fp.attributes['fingerprint'] if fp
	end
end

class Nmap::Parser::Host::OS::Port # :nodoc: all
	attr_reader :state, :proto, :num

	private

	def initialize(ports)
		parse(ports)
	end

	def parse(ports)
		@state = ports.attributes['state']
		@proto = ports.attributes['proto']
		@num = ports.attributes['portid']
	end
end

# Holds information for individual OS class records
class Nmap::Parser::Host::OS::OSClass
	attr_reader :ostype, :osgen, :osvendor, :osfamily, :accuracy

	private

	def initialize(osclass)
		parse(osclass)
	end

	def parse(osclass)
		@ostype = osclass.attributes['type']
		@osvendor = osclass.attributes['vendor']
		@osfamily = osclass.attributes['osfamily']
		@osgen = osclass.attributes['osgen']
		@accuracy = osclass.attributes['accuracy']
	end
end

# Holds information for individual OS match records
class Nmap::Parser::Host::OS::OSMatch
	attr_reader :name, :accuracy

	private

	def initialize(os)
		parse(os)
	end

	def parse(os)
		@name = os.attributes['name']
		@accuracy = os.attributes['accuracy']
	end
end

